/*
 *               ParaStation
 *
 * Copyright (C) 2002-2004 ParTec AG, Karlsruhe
 * Copyright (C) 2005 Cluster Competence Center GmbH, Munich
 *
 * $Id: ps_perf.h 5590 2008-02-18 09:47:00Z hauke $
 *
 */
/**
 * ps_perf.h: get some performance counters (eg. CPU cycles)
 *
 * @author
 *         Jens Hauke <hauke@par-tec.de>
 *
 * @file
 ***********************************************************/
#ifndef _PS_PERF_H_
#define _PS_PERF_H_


#if defined( __i386 )
/*#include <asm/msr.h>*/
#ifndef rdtsc
#define rdtsc(low,high) \
     __asm__ __volatile__("rdtsc" : "=a" (low), "=d" (high))
#endif
#ifndef rdtscl
#define rdtscl(low) \
     __asm__ __volatile__("rdtsc" : "=a" (low) : : "edx")
#endif
#ifndef rdtscll
#define rdtscll(val) \
     __asm__ __volatile__("rdtsc" : "=A" (val))
#endif


#define GET_CPU_CYCLES(/*unsigned long*/ cycl) rdtscl(cycl)
#define GET_CPU_CYCLES_LL(/*unsigned long long*/ cycl) rdtscll(cycl)

#elif defined( __ia64 )

/* #include <asm/timex.h> timex sucks in 2.6 without types.h */
static inline
unsigned long ps_get_cycles(void)
{
    unsigned long ret;
    __asm__ __volatile__ ("mov %0=ar.itc" : "=r"(ret));
    return ret;
}

#define GET_CPU_CYCLES(/*unsigned long*/ cycl) cycl = ps_get_cycles()
#define GET_CPU_CYCLES_LL(/*unsigned long long*/ cycl) cycl = ps_get_cycles()

#elif defined( __x86_64 )

#ifndef rdtsc
#define rdtsc(low,high) \
     __asm__ __volatile__("rdtsc" : "=a" (low), "=d" (high))
#endif
#ifndef rdtscl
#define rdtscl(low) \
     __asm__ __volatile__("rdtsc" : "=a" (low) : : "edx")
#endif
#ifndef rdtscll
#define rdtscll(val) \
     __asm__ __volatile__("rdtsc" : "=A" (val))
#endif


#define GET_CPU_CYCLES(/*unsigned long*/ cycl) rdtscl(cycl)
#define GET_CPU_CYCLES_LL(/*unsigned long long*/ cycl) rdtscll(cycl)

#elif defined( __powerpc64__ )

static inline unsigned long ps_get_cycles_long (void) {
  unsigned int cycles_lo, cycles_hi;
  __asm__ __volatile__("mftb %0,268\n" : "=r" (cycles_lo));
  __asm__ __volatile__("mftb %0,269\n" : "=r" (cycles_hi));
  return (((unsigned long)cycles_hi<<32)|cycles_lo);
}
static inline unsigned int ps_get_cycles (void) {
  unsigned int cycles_lo;
  __asm__ __volatile__("mftb %0,268\n" : "=r" (cycles_lo));
  return cycles_lo;
}

#define GET_CPU_CYCLES(/*unsigned long*/ cycl) cycl = ps_get_cycles()
#define GET_CPU_CYCLES_LL(/*unsigned long long*/ cycl) cycl = ps_get_cycles_long()

#elif defined( __powerpc__ )

static inline unsigned long long ps_get_cycles_long (void) {
  unsigned int cycles_lo, cycles_hi;
  __asm__ __volatile__("mftb %0,268\n" : "=r" (cycles_lo));
  __asm__ __volatile__("mftb %0,269\n" : "=r" (cycles_hi));
  return (((unsigned long long)cycles_hi<<32)|cycles_lo);
}
static inline unsigned long ps_get_cycles (void) {
  unsigned int cycles_lo;
  __asm__ __volatile__("mftb %0,268\n" : "=r" (cycles_lo));
  return cycles_lo;
}

#define GET_CPU_CYCLES(/*unsigned long*/ cycl) cycl = ps_get_cycles()
#define GET_CPU_CYCLES_LL(/*unsigned long long*/ cycl) cycl = ps_get_cycles_long()

#else

#error Unknown architecture

#endif


#endif /* _PS_PERF_H_ */
