/**************************************************** -*- linux-c -*- *
 *                  ParaStation
 *
 * Copyright (c) 2007 ParTec Cluster Competence Center GmbH
 * All rights reserved.
 **********************************************************************/
/**
 * $Id: pscom_env.c 5307 2007-10-24 14:46:02Z hauke $
 *
 * @author
 *         Jens Hauke <hauke@par-tec.de>
 *
 **********************************************************************/
#include "pscom_env.h"
#include <stdlib.h>
#include "pscom_priv.h"


void pscom_env_get_int(int *val, char *name)
{
	char *aval;

	aval = getenv(name);
	if (aval) {
		*val = atoi(aval);
		DPRINT(1, "set %s = %d", name, *val);
	} else {
		DPRINT(2, "default %s = %d", name, *val);
	}
}


void pscom_env_get_uint(unsigned int *val, char *name)
{
	char *aval;

	aval = getenv(name);
	if (aval) {
		*val = atoi(aval);
		DPRINT(1, "set %s = %u", name, *val);
	} else {
		DPRINT(2, "default %s = %u", name, *val);
	}
}


void pscom_env_get_str(char **val, char *name)
{
	char *aval;
	
	aval = getenv(name);
	if (aval) {
		*val = aval;
		DPRINT(1, "set %s = %s", name, *val);
	} else {
		DPRINT(2, "default %s = %s", name, *val);
	}
}


void pscom_env_get_dir(char **val, char *name)
{
	char *aval;
	
	aval = getenv(name);
	if (aval) {
		unsigned int len = strlen(aval);
		if (len && (aval[len-1] != '/')) {
			// append a '/'
			*val = malloc(len + 2);
			strcpy(*val, aval);
			strcat(*val,"/");
		} else {
			*val = strdup(aval);
		}
		
		DPRINT(1, "set %s = %s", name, *val);
	} else {
		DPRINT(2, "default %s = %s", name, *val);
	}
}


void pscom_env_init(void)
{
	if (!pscom.env.debug) { // only set debug once!
		pscom_env_get_int(&pscom.env.debug, ENV_DEBUG);
	}
	DPRINT(1,"# Version(PSCOM): %s", __DATE__);
	pscom_env_get_uint(&pscom.env.so_sndbuf, ENV_SO_SNDBUF);
	pscom_env_get_uint(&pscom.env.so_rcvbuf, ENV_SO_RCVBUF);
	pscom_env_get_int(&pscom.env.tcp_nodelay, ENV_TCP_NODELAY);

	// pscom_env_get_int(&env.nobgthread, ENV_NOBGTHREAD);
	pscom_env_get_int(&pscom.env.sched_yield, ENV_SCHED_YIELD);
	pscom_env_get_uint(&pscom.env.rendezvous_size, ENV_RENDEZVOUS);

	if (pscom.env.rendezvous_size != (unsigned)~0)
		pscom.env.rendezvous_size_shm = pscom.env.rendezvous_size;
	pscom_env_get_uint(&pscom.env.rendezvous_size_shm, ENV_RENDEZVOUS_SHM);

	if (pscom.env.rendezvous_size != (unsigned)~0)
		pscom.env.rendezvous_size_dapl = pscom.env.rendezvous_size;
	pscom_env_get_uint(&pscom.env.rendezvous_size_dapl, ENV_RENDEZVOUS_DAPL);

	if (pscom.env.rendezvous_size != (unsigned)~0)
		pscom.env.rendezvous_size_elan = pscom.env.rendezvous_size;
	pscom_env_get_uint(&pscom.env.rendezvous_size_elan, ENV_RENDEZVOUS_ELAN);

	pscom_env_get_int(&pscom.env.sigquit, ENV_SIGQUIT);
	pscom_env_get_uint(&pscom.env.readahead, ENV_READAHEAD);
	pscom_env_get_uint(&pscom.env.retry, ENV_RETRY);
	pscom.env.readahead = pscom_max(pscom.env.readahead, sizeof(pscom_header_net_t));

	pscom_env_get_dir(&pscom.env.plugindir, ENV_PLUGINDIR);
}

