/**************************************************** -*- linux-c -*- *
 *                  ParaStation
 *
 * Copyright (c) 2007 ParTec Cluster Competence Center GmbH
 * All rights reserved.
 **********************************************************************/
/**
 * $Id: pscom_io.h 5281 2007-10-16 15:38:57Z hauke $
 *
 * @author
 *         Jens Hauke <hauke@par-tec.de>
 *
 **********************************************************************/
#ifndef _PSCOM_IO_H_
#define _PSCOM_IO_H_

void pscom_req_free(pscom_req_t *req);

static inline
void _pscom_step(void)
{
	pscom.stat.progresscounter++;
}


/* move to state done and call io_done. unlocked version */
static inline
void pscom_req_done(pscom_req_t *req)
{
	D_TR(printf("%s(req:%p,%s)\n", __func__, req, pscom_req_state_str(req->pub.state)));

	req->pub.state |= PSCOM_REQ_STATE_IO_DONE | PSCOM_REQ_STATE_DONE;
	_pscom_step(); // ToDo: Need lock!
	
	if (req->pub.ops.io_done) {
		req->pub.ops.io_done(&req->pub);
	}
}


static inline
void _pscom_req_done(pscom_req_t *req)
{
	D_TR(printf("%s(req:%p,%s)\n", __func__, req, pscom_req_state_str(req->pub.state)));

	req->pub.state |= PSCOM_REQ_STATE_IO_DONE;
	
	_pscom_step();

	if (req->pub.ops.io_done) {
		list_add_tail(&req->next, &pscom.io_doneq);
	} else {
		req->pub.state |= PSCOM_REQ_STATE_DONE;
	}
}


/* unlocked version */
static inline
void pscom_recv_req_done(pscom_req_t *req)
{
	pscom_req_done(req);
}


/* locked version */
static inline
void _pscom_recv_req_done(pscom_req_t *req)
{
	_pscom_req_done(req);
}


static inline
void _pscom_send_req_done(pscom_req_t *req)
{
	_pscom_req_done(req);
}


static inline
void _pscom_grecv_req_done(pscom_req_t *req)
{
	D_TR(printf("%s(req:%p,%s)\n", __func__, req, pscom_req_state_str(req->pub.state)));
	// assert(!genreq->rendezvous_req);
	req->pub.state |= PSCOM_REQ_STATE_DONE;
	_pscom_step();
	pscom_req_free(req);
}


static inline 
void _genreq_enq(pscom_con_t *con, pscom_req_t *genreq)
{
	pscom_sock_t *sock = get_sock(con->pub.socket);
	list_add_tail(&genreq->next, &con->genrecvq);
	list_add_tail(&genreq->gen_next_any, &sock->genrecvq_any);
}


static inline
void _genreq_dequeue(pscom_req_t *genreq)
{
	list_del(&genreq->next);
	list_del(&genreq->gen_next_any);
}



#endif /* _PSCOM_IO_H_ */
