/*
 *               ParaStation
 *
 * Copyright (C) 2007 Cluster Competence Center GmbH, Munich
 *
 * $Id: psoib.h 5248 2007-10-02 12:19:10Z hauke $
 *
 * psoib.c: OPENIB/Infiniband communication
 *
 * @author
 *         Jens Hauke <hauke@par-tec.de>
 *
 * @file
 ***********************************************************/
#ifndef _PSOIB_H_
#define _PSOIB_H_

#include <stdint.h>
#include <sys/uio.h>

typedef struct psoib_con_info psoib_con_info_t;
typedef struct hca_info hca_info_t;
typedef struct port_info port_info_t;

// contact endpoint info
typedef struct psoib_info_msg_s {
	uint16_t	lid;
	uint32_t	qp_num;  /* QP number */
	void		*remote_ptr; /* Info about receive buffers */
	uint32_t	remote_rkey;
} psoib_info_msg_t;


int psoib_init(void);


// Connection handling:

psoib_con_info_t *psoib_con_create(void);
void	psoib_con_free(psoib_con_info_t *con_info);

int	psoib_con_init(psoib_con_info_t *con_info, hca_info_t *hca_info, port_info_t *port_info);
int	psoib_con_connect(psoib_con_info_t *con_info, psoib_info_msg_t *info_msg);
void	psoib_con_cleanup(psoib_con_info_t *con_info, hca_info_t *hca_info);

void	psoib_con_get_info_msg(psoib_con_info_t *con_info /* in */, psoib_info_msg_t *info /* out */);


/* returnvalue like read() , except on error errno is negative return */
int psoib_recvlook(psoib_con_info_t *con_info, void **buf);
void psoib_recvdone(psoib_con_info_t *con_info);


/* returnvalue like write(), except on error errno is negative return */
/* It's important, that the sending side is aligned to IB_MTU_SPEC,
   else we loose a lot of performance!!! */
int psoib_sendv(psoib_con_info_t *con_info, struct iovec *iov, int size);
void psoib_send_eof(psoib_con_info_t *con_info);

extern int psoib_debug;

#endif /* _PSOIB_H_ */
